//
// GameX - Sound- and Music-related GameXClass Class Code
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 
// this is a version of GameX that has been modified (by Justin Pease)


// KNOWN ISSUES:

// when resuming MIDI music that has been paused, oftentimes a single note will play for quite a few seconds
// over the rest of the music which is continuing as normal.
// this is due to the tricky use of tempo change to pause the music,
// a better solution would be to remember where it is and stop it, then start again at that point on resume,
// but this requires making use of DirectMusic 8.0 interface for PlayEx()
// (which *is* possible to do while remaining compatible with computers that only have DX7)

#include "gamex-win-dx.h"


// DirectSound
#ifndef DIRECTSOUND_VERSION 
#define DIRECTSOUND_VERSION 0x0700 // specify version 7.0 of DirectSound
//#include <dsound.h>
#endif

// DirectMusic
#ifndef DIRECTMUSIC_VERSION
#define DIRECTMUSIC_VERSION 0x0700 // specify version 7.0 of DirectMusic
//#include <dmksctrl.h>
#include <dmusici.h>
//#include <dmusicc.h>
#include <dmusicf.h>
#endif

// DirectShow
#ifndef DIRECTSHOW_VERSION
#define DIRECTSHOW_VERSION 0x0700 // specify version 7.0 of DirectShow
#include <strmif.h>
#include <control.h>
#include <uuids.h>
#include <evcode.h>
#endif

#include "gamex-win-dx.h"
#include "gamex-sound.h"

// vol: 0.0 means silent, 1.0 means loud, etc.
// pan: -1.0 means left only, 0.0 means center (both), 1.0 means right only, 1.5 means right only half volume, -1.5 means left only half volume, <=-2.0 and >=2.0 mean no volume, etc.
// freq: 1.0 means normal frequency, 2.0 means twice as high-pitched, 0.5 means half pitch, etc.
void GameXClass::PlaySound (SoundX * snd, SoundPlayMode mode, float vol, float pan, float freq)
{
	assertMsg2(z_win_hwnd, "Trying to play the sound \"%s\" before GameX has been initialized.", snd ? snd->GetFilename() : "(null)");
	if(z_sound_support)
	{
		if(snd->NeedUpdate() == true) z_UpdateSound((SoundXPtr)snd) ;	
		LPDIRECTSOUNDBUFFER buffer = snd->GetDSBuffer();
		if(buffer == NULL) return;
		if(pan > 1.0f)
		{
			vol *= max(0.0f, 1.0f - (pan-1.0f));
			pan = 1.0f;
		} else if(pan < -1.0f)
		{
			vol *= max(0.0f, 1.0f - (-pan-1.0f));
			pan = -1.0f;
		}
		if(vol > 1.0f) vol = 1.0f; // DirectSound does not support sound amplification (volume > 100%)
		if(vol < 0.0f) vol = 0.0f;
		if(mode & PLAY_NOTIFLESSVOL)
		{
			DWORD status;
			buffer->GetStatus(&status);
			if(status & DSBSTATUS_PLAYING)
			{
				long old_vol;
				buffer->GetVolume(&old_vol);
				if(z_ToAttenuation(vol) < old_vol)
					return;
			}
		}
		if(mode & PLAY_REWIND) buffer->SetCurrentPosition(0) ;

		buffer->SetVolume(z_ToAttenuation(vol)) ;
		buffer->SetPan((long) (pan*2048.0)) ;
		buffer->SetFrequency((DWORD)(snd->GetSamplesPerSec()*freq)) ;
		buffer->Play(0,0,(mode & PLAY_LOOP ? DSBPLAY_LOOPING : 0)) ;
	}
}

bool GameXClass::IsSoundPlaying(const SoundX * snd)
{
	if(snd->NeedUpdate() == true) z_UpdateSound(snd) ;	
	const LPDIRECTSOUNDBUFFER buffer = snd->GetDSBuffer();
	if(buffer == NULL) return false;

	DWORD status;
	buffer->GetStatus(&status);
	if(status & DSBSTATUS_PLAYING)
		return true;
	else
		return false;
}

void GameXClass::RewindSound (SoundX * snd)
{
	if(z_sound_support)
	{
		if(snd->NeedUpdate() == true) z_UpdateSound (snd) ;	
		snd->GetDSBuffer()->SetCurrentPosition(0) ;
	}
}


void GameXClass::StopSound (SoundX * snd)
{
	if(z_sound_support)
	{
		snd->GetDSBuffer()->Stop() ;
	}
}

void GameXClass::z_UpdateSound (const SoundX * snd) 
{
	short *buf_a, *buf_b;	
	unsigned long bytes_a, bytes_b; 

	int z_num_samples, z_num_channels;
	int result;
	int n, c;
	
	// Determine Direct Sound buffer to use based on SoundX object.
	LPDIRECTSOUNDBUFFER buffer = snd->GetDSBuffer();

	if(buffer == NULL)
	{
		snd->PrepareDSBuffer();
		buffer = snd->GetDSBuffer();
		z_MemoryRecordChange(snd->GetNumBytes());
	}

	if(buffer == NULL) return;

	// Get requested Direct Sound buffer capabilities
	DSBCAPS caps;
	ZeroMemory(&caps, sizeof(caps)) ;
	caps.dwSize = sizeof(caps) ;
	if(DS_OK != buffer->GetCaps (&caps)) z_ReportProblem(GAMEX_SOUNDCAPS_FAILED) ;
		
	// Get and check number of samples (SoundX sample count should match DS buffer count)
	z_num_channels = snd->GetNumChannels() ;
	z_num_samples = caps.dwBufferBytes / z_num_channels;

	if(z_num_samples != snd->GetNumSamples()) z_num_samples = snd->GetNumSamples();
		
	
	// Lock Direct Sound buffer
	result = buffer->Lock (0, caps.dwBufferBytes, (void**) &buf_a, &bytes_a, (void**) &buf_b, &bytes_b, 0) ;
	n = 0;
	while(result == DSERR_BUFFERLOST && n<10)
	{
		buffer->Restore() ;
		result = buffer->Lock (0, caps.dwBufferBytes, (void**) &buf_a, &bytes_a, (void**) &buf_b, &bytes_b, 0) ;
		n++;
	}
	// Check for possible buffer errors
	if(DS_OK != result) {		
		if(result == DSERR_BUFFERLOST)			z_win_status = GAMEX_SOUNDLOST_FAIL;	
		else if(result == DSERR_INVALIDCALL) 	z_win_status = GAMEX_SOUNDINVALID_FAIL;
		else if(result == DSERR_INVALIDPARAM)	z_win_status = GAMEX_SOUNDPARAM_FAIL;
		else if(result == DSERR_PRIOLEVELNEEDED) z_win_status = GAMEX_SOUNDPRIOLEVEL_FAIL;
		z_ReportProblem() ;
	}
	// Copy SoundX data to Direct Sound buffer
	if(snd->GetBPS() <= 8)
	{
		unsigned char *data = (unsigned char*) snd->GetData() ;
		unsigned char *curr_a = (unsigned char*) buf_a;
		for(n = 0; n < z_num_samples; n++)
			for(c = 0; c < z_num_channels; c++)
				*curr_a++ = *data++;
	} else
	{
		short *data = (short*) snd->GetData() ;
		short *curr_a = buf_a;
		curr_a = buf_a;
		for(n = 0; n < z_num_samples; n++)
			for(c = 0; c < z_num_channels; c++) 
				*curr_a++ = *data++;			
	}
	
	// Unlock Direct Sound buffer
	buffer->Unlock (&buf_a, bytes_a, &buf_b, bytes_b) ;

	snd->SetRefresh (false) ;
}






// Plays a MusicX object, and immediately stops playing any other music that may be playing
// times of 0 means infinite loop, 1 means play once, 2 means play twice, etc.
// speed_factor of 1.0 means normal speed, 2.0 means double speed, etc.
void GameXClass::PlayMusic (MusicX * music, int times, float volume, float fade_in_seconds, float tempo, float pitch)
{
	StopMusic(z_current_music,0.0f);
	z_current_music = music;
	if(music == NULL) return;
	music->playing_loops_left = times ? times : 32767;
	switch(music->format)
	{
		case MEDIA_TYPE_DMUSIC:
		{
			if(music->GetSegment() == NULL)
			{
				if(z_win_log_debug)
					GameX.z_SafeMessageBox("Error: Attempting to play a music file that did not load properly.\nMake sure your music files are in the proper location and in .mid format.\nAlso make sure any music files needed by the game are not open in another program.", "GameX Error",MB_OK|MB_ICONSTOP) ;

				return;
			}

			GameX.z_InitDirectMusicPerformance();

			VolumeFadeMusic(music, 0.0f, 0.0f);

			if(music->GetPlayState()) { // if already playing, have to wait for change to take effect // HACK
				int clocker = clock();
				while(clock() - clocker < 80) {}
			}

			if(times) times--;
			else times = DMUS_SEG_REPEAT_INFINITE;
			music->GetSegment()->SetRepeats(times);
			SetMusicTempo(tempo);
			IDirectMusicSegmentState * idmss = music->GetSegmentState();
			z_dm_performance->PlaySegment(music->GetSegment(), 0, 0, &idmss);
		}	break;
		case MEDIA_TYPE_DSHOW:
		{
			HRESULT status;
			try
			{
				status = music->dsh_mc->Run();
			}
			catch(...)
			{
				status = -1;
			}

			if(FAILED(status))
			{
				if(z_win_err_debug)
					debug.Output("GameX: Failed to Run WMA or MP3\n");
				return;
			}
		}	break;
		case MEDIA_TYPE_NONE:
			return;
	}

	PitchBendMusic(music, pitch,0.0f);
	VolumeFadeMusic(music, volume, fade_in_seconds);
	music->SetPlayState(1);
}

// Fades the volume to 0 over given seconds (if seconds>0), then stops the music from playing
// If it's an MP3, fade not supported yet so acts as if seconds = 0.
void GameXClass::StopMusic (MusicX * music, float seconds)
{
	if(music == NULL) music = z_current_music;
	if(music == NULL) return;
	switch(music->format)
	{
		case MEDIA_TYPE_DMUSIC:
		{
			if(music->GetSegment() == NULL) return;
			VolumeFadeMusic(music, 0.0f, seconds); // fade for given seconds

			// calculate time until after fade
			DMUS_TEMPO_PARAM tempo;
			music->GetSegment()->GetParam(GUID_TempoParam, 0xFFFF, 0, 0, NULL, &tempo);
			seconds *= (float)(tempo.dblTempo * 10.0);
			MUSIC_TIME musictime;
			z_dm_performance->GetTime(NULL, &musictime);
			musictime += (MUSIC_TIME)seconds;

			z_dm_performance->Stop(music->GetSegment(), NULL, musictime, 0); // stop after fade
			music->SetPlayState(seconds?2:0);
		}	break;
		case MEDIA_TYPE_DSHOW:
		{
			music->dsh_mc->Stop();
			music->dsh_mp->put_CurrentPosition(0);
			music->SetPlayState(0);
			if(music->dsh_vw)
				music->dsh_vw->put_Visible(false);
		}	break;
	}
	z_current_music = NULL;
}

// Fades the music up or down to the given volume factor, over the given time
// seconds is how many seconds to alter volume over
// target_factor of 0.0f is fade to silence, 0.5f is fade to half master volume, 1.0f is no change, 1.27f is fade to max amplified
void GameXClass::VolumeFadeMusic(MusicX * music, float volume_factor, float seconds)
{
	switch(music->format)
{
	case MEDIA_TYPE_DMUSIC:
		if(seconds < 0.05f) seconds = 0.05f;
		music->Curve((long)(1000*seconds), (int)max(-20,min(127, 100*volume_factor)), DMUS_CURVET_CCCURVE, DMUS_CURVES_LINEAR, (int)max(-20,min(127, 100*music->GetVolume())));
		music->RememberVol(volume_factor);
		break;
	case MEDIA_TYPE_DSHOW:
		music->dsh_ba->put_Volume(z_ToAttenuation(volume_factor*z_master_music_volume));
		break;
	}
}

// Bends the pitch of the mucis up or down to the given pitch factor, over the given time
// seconds is how many seconds to alter pitch over. This method does not change speed of play
// pitch_factor of 0.5f is half pitch (low frequency), 2.0f is double pitch (high frequency), 1.0f is no change, etc.
void GameXClass::PitchBendMusic(MusicX * music, float pitch_factor, float seconds)
{
	switch(music->format)
{
	case MEDIA_TYPE_DMUSIC:
		if(seconds < 0.05f) seconds = 0.05f;
		music->Curve((long)(1000*seconds), (int)max(1000, min(32767, 8000*pitch_factor)), DMUS_CURVET_PBCURVE, DMUS_CURVES_LINEAR);
		break;
	}
}

// Sets the tempo, or speed of play, of all music -- does not change the pitch
// 0.5f is half speed, 1.0f is normal speed, 2.0f is double speed, etc.
// Can be called before or after you start playing the song
// note: there may be a small time delay after calling this function before the change takes effect
void GameXClass::SetMusicTempo(float tempo_factor)
{
	tempo_factor = min(DMUS_MASTERTEMPO_MAX, max(DMUS_MASTERTEMPO_MIN, tempo_factor));
	z_dm_performance->SetGlobalParam(GUID_PerfMasterTempo, &tempo_factor, sizeof(float));
	z_master_music_tempo = tempo_factor;
}

// Sets the volume of all music
// Can be called before or after you start playing a song
// If an individual music song has its volume faded to 50% and the master volume is 50%
// that will result in the music being played at 25% volume
void GameXClass::SetMasterMusicVolume(float volume_factor)
{
	long volume = z_ToAttenuation(volume_factor);
	z_dm_performance->SetGlobalParam(GUID_PerfMasterVolume, &volume, sizeof(long));
	z_master_music_volume = volume_factor;
}

// returns true if the given song is still playing, looping, or paused, false otherwise
bool GameXClass::IsMusicPlaying(const MusicX * music)
{
	if(music == NULL)
		return false;

	switch(music->format)
	{
		case MEDIA_TYPE_DMUSIC:
			return (z_dm_performance && z_dm_performance->IsPlaying(music->GetSegment(),NULL) == S_OK);

		case MEDIA_TYPE_DSHOW:
			if(z_current_music == NULL)
				return false;
			long lEvent;
			music->dsh_me->WaitForCompletion(0,&lEvent); // wait time = 0, just to get playing status
			return (lEvent != EC_COMPLETE);
	}

	return false;
}


// pauses whatever song is currently playing (has only been tested internally to GameX)
void GameXClass::PauseMusic ()
{
	if(z_current_music == NULL) return;
	if(z_current_music->GetPlayState() != 1) return;

	switch(z_current_music->format)
	{
		case MEDIA_TYPE_DMUSIC:
		{
			float current_tempo = z_master_music_tempo;
			SetMusicTempo(DMUS_MASTERTEMPO_MIN*2);
			z_master_music_tempo = current_tempo;

			float current_volume = z_master_music_volume;
			SetMasterMusicVolume(0);
			z_master_music_volume = current_volume;
		}	break;
		case MEDIA_TYPE_DSHOW:
		{
			z_current_music->dsh_mc->Pause();
		}	break;
	}
	z_current_music->SetPlayState(2);
}

// resumes whatever song is currently playing (has only been tested internally to GameX)
// due to the way it is paused, a stray note might play for a while after resuming MIDIs
void GameXClass::ResumeMusic ()
{
	if(z_current_music == NULL) return;

	switch(z_current_music->format)
	{
		case MEDIA_TYPE_DMUSIC:
		{
			SetMusicTempo(z_master_music_tempo);
			SetMasterMusicVolume(z_master_music_volume);
		}	break;
		case MEDIA_TYPE_DSHOW:
		{
			if(z_current_music->GetPlayState() != 0)
			{
				z_current_music->dsh_mc->Run();
				z_current_music->AdjustSize();
			}
		}	break;
	}
	z_current_music->SetPlayState(1);
}


void GameXClass::z_HandleDirectShowEvent()
{
	if(z_current_music == NULL) return;

	long evCode, param1, param2;
	HRESULT status;
	do
	{
		status = z_current_music->dsh_me->GetEvent(&evCode, &param1, &param2, 0);
		if(SUCCEEDED(status))
		{
			status = z_current_music->dsh_me->FreeEventParams(evCode, param1, param2);
			if(SUCCEEDED(status)) { 
				switch(evCode)
			{
				case EC_COMPLETE:
					if(z_current_music->playing_loops_left)
						z_current_music->playing_loops_left--;
					else
					{
						z_current_music->dsh_mc->Stop();
						if(z_current_music->dsh_vw)
							z_current_music->dsh_vw->put_Visible(false);
					}

					z_current_music->dsh_mp->put_CurrentPosition(0);
					break;
				case EC_USERABORT:
					z_current_music->dsh_mc->Stop();
					if(z_current_music->dsh_vw)
						z_current_music->dsh_vw->put_Visible(false);
					break;
				}
			}
		}
	} while(SUCCEEDED(status));
}
